from __future__ import annotations

import os
from pprint import pprint

from geocondense._core import dissect_geojson as dissect_geojson_impl


def dissect_geojson(
    *,
    input_path: str,
    output_geometry: str | None = None,
    output_properties: str | None = None,
    output_observations: str | None = None,
    output_others: str | None = None,
    indent: bool = False,
):
    # mkdir -p
    for path in [
        output_geometry,
        output_properties,
        output_observations,
        output_others,
    ]:
        if not path:
            continue
        path = os.path.abspath(path)
        os.makedirs(os.path.dirname(path), exist_ok=True)

    # call c++ dissect_geojson
    succ = dissect_geojson_impl(
        input_path=input_path,
        output_geometry=output_geometry,
        output_properties=output_properties,
        output_observations=output_observations,
        output_others=output_others,
        indent=indent,
    )
    if not succ:
        pprint(locals())  # noqa: T203
        msg = f"failed to dissect geojson: {input_path}"
        raise Exception(msg)


if __name__ == "__main__":
    import fire

    fire.core.Display = lambda lines, out: print(*lines, file=out)
    fire.Fire(dissect_geojson)



from __future__ import annotations

import json
import os
from itertools import chain

import numpy as np
import open3d as o3d
from concave_hull import concave_hull_indexes
from loguru import logger
from polyline_ruler import douglas_simplify_mask, tf


def condense_pointcloud_impl(
    *,
    pcd: o3d.geometry.PointCloud,
    output_fence_path: str | None = None,
    output_grids_dir: str | None = None,
    grid_resolution: float = 0.0001,
    compress_pcd: bool = False,
):
    assert (
        output_fence_path or output_grids_dir
    ), "should specify either --output_fence_path or --output_grids_dir"
    if output_fence_path:
        assert output_fence_path.endswith(
            ".json"
        ), f"invalid voxel dump: {output_fence_path}, should be a json file"
    wgs84_scale = 1 / grid_resolution
    assert wgs84_scale == int(wgs84_scale), f"bad grid_resolution: {grid_resolution}"
    wgs84_scale = int(wgs84_scale)

    ecefs = np.asarray(pcd.points)
    assert len(ecefs), "not any points in pointcloud"
    R = np.linalg.norm(ecefs[0])
    assert (
        R > 6300 * 1000
    ), f"data (should be in ECEF) not on earth? (forgot to specify --center?), R is: {R}"
    anchor = tf.ecef2lla(*ecefs[0])
    anchor[:2] = [round(x * wgs84_scale) / wgs84_scale for x in anchor[:2]]
    anchor[2] = 0.0
    anchor_text = "_".join([str(x) for x in anchor])

    T_ecef_enu = tf.T_ecef_enu(*anchor)
    pcd.transform(np.linalg.inv(T_ecef_enu))
    enus = np.asarray(pcd.points)
    rgbs = np.asarray(pcd.colors)
    pmin = pcd.get_min_bound()
    pmax = pcd.get_max_bound()
    lla_bounds = tf.enu2lla(
        [pmin - 10.0, pmax + 10.0],
        anchor_lla=anchor,
        cheap_ruler=False,
    )
    lon0, lat0 = lla_bounds[0][:2]
    lon1, lat1 = lla_bounds[1][:2]
    lon0, lon1 = (round(x * wgs84_scale) / wgs84_scale for x in [lon0, lon1])
    lat0, lat1 = (round(x * wgs84_scale) / wgs84_scale for x in [lat0, lat1])
    lons = np.arange(
        lon0 - grid_resolution, lon1 + grid_resolution + 1e-15, grid_resolution
    )
    lats = np.arange(
        lat0 - grid_resolution, lat1 + grid_resolution + 1e-15, grid_resolution
    )
    lons = [round(x * wgs84_scale) / wgs84_scale for x in lons]
    lats = [round(x * wgs84_scale) / wgs84_scale for x in lats]
    assert len(lons) > 1
    assert len(lats) > 1
    xs = tf.lla2enu(
        [[x, lats[0], 0.0] for x in lons],
        anchor_lla=anchor,
        cheap_ruler=False,
    )[:, 0]
    ys = tf.lla2enu(
        [[lons[0], x, 0.0] for x in lats],
        anchor_lla=anchor,
        cheap_ruler=False,
    )[:, 1]

    pcd1, _, idxes = pcd.voxel_down_sample_and_trace(
        1.0,
        min_bound=pmin,
        max_bound=pmax,
    )
    idxes = [np.array(i) for i in idxes]
    xyzs = np.asarray(pcd1.points)

    # export voxels
    points = np.array(
        [
            *xyzs,
            *(xyzs + [+0.1, 0, 0]),
            *(xyzs + [-0.1, 0, 0]),
            *(xyzs + [0, +0.1, 0]),
            *(xyzs + [0, -0.1, 0]),
        ]
    )
    concave_hull = concave_hull_indexes(
        points[:, :2],
        length_threshold=2.0,
    )
    concave_hull = [*concave_hull, concave_hull[0]]
    llas = tf.enu2lla(points[concave_hull], anchor_lla=anchor)
    llas[:, :2] = llas[:, :2].round(6)
    llas[:, 2] = llas[:, 2].round(1)
    mask = douglas_simplify_mask(tf.lla2enu(llas), epsilon=0.5).astype(bool)
    llas = llas[mask]
    os.makedirs(os.path.dirname(os.path.abspath(output_fence_path)), exist_ok=True)
    with open(output_fence_path, "w") as f:
        logger.info(f"writing to {output_fence_path}")
        json.dump(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "geometry": {
                            "type": "Polygon",
                            "coordinates": [llas.tolist()],
                        },
                        "properties": {
                            "type": "pointcloud",
                            "#points": len(enus),
                            "#voxels_1m^3": len(xyzs),
                            "bbox": lla_bounds.tolist(),
                            "size": (pmax - pmin).round(2).tolist(),
                        },
                    }
                ],
            },
            f,
            indent=4,
        )

    if not output_grids_dir:
        return
    output_grids_dir = os.path.abspath(output_grids_dir)
    os.makedirs(output_grids_dir, exist_ok=True)
    for ii, (x0, x1) in enumerate(zip(xs[:-1], xs[1:])):
        for jj, (y0, y1) in enumerate(zip(ys[:-1], ys[1:])):
            mask = np.logical_and(
                np.logical_and(x0 <= xyzs[:, 0], xyzs[:, 0] < x1),
                np.logical_and(y0 <= xyzs[:, 1], xyzs[:, 1] < y1),
            )
            if not np.any(mask):
                continue
            related = [idxes[i] for i in np.where(mask)[0]]
            related = sorted(chain.from_iterable(related))
            grid = o3d.geometry.PointCloud()
            grid.points = o3d.utility.Vector3dVector(enus[related])
            grid.colors = o3d.utility.Vector3dVector(rgbs[related])
            bounds = lons[ii], lats[jj], lons[ii + 1], lats[jj + 1]
            bounds = "_".join([str(x) for x in bounds])
            path = f"{output_grids_dir}/grid_res{grid_resolution}_{bounds}_anchor_{anchor_text}_raw.pcd"
            logger.info(f"writing #{len(grid.points):,} points to {path}")
            assert o3d.io.write_point_cloud(
                path,
                grid,
                compressed=compress_pcd,
            ), f"failed to dump grid pcd to {path}"
            for res in [0.25, 0.05]:
                path = f"{output_grids_dir}/grid_res{grid_resolution}_{bounds}_anchor_{anchor_text}_voxel{res}.pcd"
                small_grid = grid.voxel_down_sample(res)
                logger.info(f"writing #{len(small_grid.points):,} points to {path}")
                assert o3d.io.write_point_cloud(
                    path,
                    small_grid,
                    compressed=compress_pcd,
                ), f"failed to dump grid pcd to {path}"


def condense_pointcloud(
    *,
    input_path: str,
    output_fence_path: str | None = None,
    output_grids_dir: str | None = None,
    grid_resolution: float = 0.0001,
    compress_pcd: bool = False,
    center: tuple[float, float, float] | None = None,
):
    assert (
        output_fence_path or output_grids_dir
    ), "should specify either --output_fence_path or --output_grids_dir"
    pcd = o3d.io.read_point_cloud(input_path)
    if center:
        if isinstance(center, str):
            center = [float(x) for x in center.split(",")]
        assert isinstance(center, (list, tuple)), f"invalid center: {center}"
        lon, lat, alt = (*center, 0.0) if len(center) == 2 else tuple(center)
        pcd.transform(tf.T_ecef_enu(lon=lon, lat=lat, alt=alt))
    return condense_pointcloud_impl(
        pcd=pcd,
        output_fence_path=output_fence_path,
        output_grids_dir=output_grids_dir,
        grid_resolution=grid_resolution,
        compress_pcd=compress_pcd,
    )


if __name__ == "__main__":
    import fire

    fire.core.Display = lambda lines, out: print(*lines, file=out)
    fire.Fire(condense_pointcloud)



from __future__ import annotations
import typing

__all__ = ["CondenseOptions", "condense_geojson", "dissect_geojson"]

class CondenseOptions:
    def __init__(self) -> None:
        """
        Default constructor for CondenseOptions
        """
    @property
    def debug(self) -> bool:
        """
        Debug option
        """
    @debug.setter
    def debug(self, arg0: bool) -> None: ...
    @property
    def douglas_epsilon(self) -> float:
        """
        Epsilon value for Douglas-Peucker algorithm
        """
    @douglas_epsilon.setter
    def douglas_epsilon(self, arg0: float) -> None: ...
    @property
    def grid_features_keep_properties(self) -> bool:
        """
        Option to keep properties for grid features
        """
    @grid_features_keep_properties.setter
    def grid_features_keep_properties(self, arg0: bool) -> None: ...
    @property
    def grid_h3_resolution(self) -> int:
        """
        H3 resolution for grid features
        """
    @grid_h3_resolution.setter
    def grid_h3_resolution(self, arg0: int) -> None: ...
    @property
    def indent(self) -> bool:
        """
        Indentation option for JSON output
        """
    @indent.setter
    def indent(self, arg0: bool) -> None: ...
    @property
    def sort_keys(self) -> bool:
        """
        Option to sort keys in JSON output
        """
    @sort_keys.setter
    def sort_keys(self, arg0: bool) -> None: ...
    @property
    def sparsify_h3_resolution(self) -> int:
        """
        H3 resolution for sparsification
        """
    @sparsify_h3_resolution.setter
    def sparsify_h3_resolution(self, arg0: int) -> None: ...
    @property
    def sparsify_upper_limit(self) -> int:
        """
        Upper limit for sparsification
        """
    @sparsify_upper_limit.setter
    def sparsify_upper_limit(self, arg0: int) -> None: ...

def condense_geojson(
    *,
    input_path: str,
    output_index_path: str | None = None,
    output_strip_path: str | None = None,
    output_grids_dir: str | None = None,
    options: CondenseOptions = ...,
) -> bool:
    """
    Condense GeoJSON data.

    Args:
        input_path: Path to the input GeoJSON file.
        output_index_path: Optional path for the output index file.
        output_strip_path: Optional path for the output strip file.
        output_grids_dir: Optional directory for output grid files.
        options: CondenseOptions object with configuration options.

    Returns:
        bool: True if the operation was successful, False otherwise.
    """

def dissect_geojson(
    *,
    input_path: str,
    output_geometry: str | None = None,
    output_properties: str | None = None,
    output_observations: str | None = None,
    output_others: str | None = None,
    indent: bool = False,
) -> bool:
    """
    Dissect GeoJSON data into separate components.

    Args:
        input_path: Path to the input GeoJSON file.
        output_geometry: Optional path for the output geometry file.
        output_properties: Optional path for the output properties file.
        output_observations: Optional path for the output observations file.
        output_others: Optional path for other output data.
        indent: Boolean flag to enable indentation in output JSON.

    Returns:
        bool: True if the operation was successful, False otherwise.
    """

__version__: str = "0.0.3"



from __future__ import annotations

import argparse
import glob
import json
import os
from collections import defaultdict
from pathlib import Path

import open3d as o3d
from loguru import logger
from polyline_ruler import tf

from geocondense.condense_geojson import condense_geojson
from geocondense.condense_pointcloud import condense_pointcloud_impl
from geocondense.dissect_geojson import dissect_geojson
from geocondense.utils import md5sum as default_md5sum
from geocondense.utils import read_json, write_json


def resolve_center(
    center: str | tuple[float, float, float] | None,
) -> tuple[float, float, float] | None:
    if not center:
        return None
    if isinstance(center, str):
        center = [float(x) for x in center.split(",")]
    assert isinstance(center, (list, tuple)), f"invalid center: {center}"
    lon, lat, alt = (*center, 0.0) if len(center) == 2 else tuple(center)
    return lon, lat, alt


def default_handle_semantic(
    path: str,
    *,
    workdir: str,
    uuid: str,
    center: tuple[float, float, float] | None,
) -> tuple[str, str]:
    dissect_input_path = path
    condense_input_path = path
    return dissect_input_path, condense_input_path


def default_handle_pointcloud(
    path: str,
    *,
    workdir: str,
    uuid: str,
    center: tuple[float, float, float] | None,
) -> o3d.geometry.PointCloud:
    pcd = o3d.io.read_point_cloud(path)
    if center:
        lon, lat, alt = center
        pcd.transform(tf.T_ecef_enu(lon=lon, lat=lat, alt=alt))
    return pcd


def condense_semantic(
    dissect_input_path: str,
    condense_input_path: str,
    *,
    output_dir: str,
) -> str:
    dissect_geojson(
        input_path=dissect_input_path,
        output_geometry=f"{output_dir}/geometry.json",
        output_properties=f"{output_dir}/properties.json",
        output_observations=f"{output_dir}/observations.json",
        output_others=f"{output_dir}/others.json",
        indent=True,
    )
    condense_geojson(
        input_path=condense_input_path,
        output_index_path=f"{output_dir}/meta.json",
        output_strip_path=f"{output_dir}/main.json",
        output_grids_dir=f"{output_dir}/grids",
        indent=True,
    )
    path = f"{output_dir}/index.json"
    files = sorted(
        os.path.basename(p) for p in glob.glob(f"{output_dir}/grids/h3_cell_*.json")
    )
    grids = defaultdict(dict)  # res -> h3idx -> file
    for p in files:
        # e.g. h3_cell_8_08831aa4301fffff.json
        res, idx = p.rsplit(".", 1)[0].split("_")[2:]
        grids[res][idx] = f"grids/{p}"
    write_json(
        path,
        {
            "type": "semantic",
            "main": "main.json",
            "grids": grids,
            "meta": "meta.json",
            "geometry": "geometry.json",
            "properties": "properties.json",
            "observations": "observations.json",
            "others": "others.json",
        },
    )
    Path(f"{output_dir}.semantic").touch()
    return path


def condense_pointcloud(
    pcd: o3d.geometry.PointCloud,
    *,
    output_dir: str,
) -> str:
    condense_pointcloud_impl(
        pcd=pcd,
        output_fence_path=f"{output_dir}/main.json",
        output_grids_dir=f"{output_dir}/grids",
        grid_resolution=0.0001,
    )
    path = f"{output_dir}/index.json"
    files = sorted(
        os.path.basename(p) for p in glob.glob(f"{output_dir}/grids/grid_res*.pcd")
    )
    grids = defaultdict(
        defaultdict(defaultdict(dict).copy).copy
    )  # res -> bbox -> category -> {info}
    for p in files:
        # grid_res0.0001_116.311_39.8959_116.3111_39.896_anchor_116.3114_39.8958_0.0_raw.pcd
        # grid_res0.0001_116.311_39.8958_116.3111_39.8959_anchor_116.3114_39.8958_0.0_voxel0.25.pcd
        res, lon0, lat0, lon1, lat1, _, ax, ay, az, category = p[8:-4].split("_")
        grids[res][f"{lon0},{lat0},{lon1},{lat1}"][category] = {
            "file": f"grids/{p}",
            "anchor": [float(ax), float(ay), float(az)],
        }
    write_json(
        path,
        {
            "type": "pointcloud",
            "main": "main.json",
            "grids": grids,
        },
    )
    Path(f"{output_dir}.pointcloud").touch()
    return path


@logger.catch(reraise=True)
def condense(
    *,
    workdir: str,
    semantic_files: list[str] = None,
    pointcloud_files: list[str] = None,
    center: tuple[float, float, float] | None = None,
    # handlers
    handle_semantic=default_handle_semantic,
    handle_pointcloud=default_handle_pointcloud,
    md5sum=default_md5sum,
) -> dict[str, str]:
    assert not (
        semantic_files is None and pointcloud_files is None
    ), "should specify either --semantic_files or --pointcloud_files (or both)"
    semantic_files = semantic_files or []
    pointcloud_files = pointcloud_files or []
    logger.info(f"semantic files: {semantic_files} (#{len(semantic_files)})")
    logger.info(f"pointcloud files: {pointcloud_files} (#{len(pointcloud_files)})")

    for p in [*semantic_files, *pointcloud_files]:
        assert os.path.isfile(p), f"{p} does not exist"
    center = resolve_center(center)
    os.makedirs(os.path.abspath(workdir), exist_ok=True)
    index_map = {}
    for path in semantic_files:
        uuid = md5sum(path)
        odir = f"{workdir}/{uuid}"
        index = f"{odir}/index.json"
        if os.path.isfile(index):
            logger.info(f"skip condensing {path}, index exists: {index}")
            index_map[path] = {"uuid": uuid, **read_json(index)}
            continue
        dissect_input, condense_input = handle_semantic(
            path,
            workdir=workdir,
            uuid=uuid,
            center=center,
        )
        index = condense_semantic(dissect_input, condense_input, output_dir=odir)
        index_map[path] = {"uuid": uuid, **read_json(index)}
    for path in pointcloud_files:
        uuid = md5sum(path)
        odir = f"{workdir}/{uuid}"
        index = f"{odir}/index.json"
        if os.path.isfile(index):
            logger.info(f"skip condensing {path}, index exists: {index}")
            index_map[path] = {"uuid": uuid, **read_json(index)}
            continue
        pcd = handle_pointcloud(
            path,
            workdir=workdir,
            uuid=uuid,
            center=center,
        )
        index = condense_pointcloud(pcd, output_dir=odir)
        index_map[path] = {"uuid": uuid, **read_json(index)}
    return index_map


def main(
    handle_semantic=default_handle_semantic,
    handle_pointcloud=default_handle_pointcloud,
):
    prog = "python3 -m geocondense.condense"
    description = "condense semantic & pointcloud"
    parser = argparse.ArgumentParser(prog=prog, description=description)
    parser.add_argument(
        "workdir",
        type=str,
        help="workdir",
    )
    parser.add_argument(
        "--center",
        type=str,
        help='specify with --center="lon,lat,alt" or --center="path/to/center.txt"',
    )
    parser.add_argument(
        "--semantic_files",
        nargs="*",
        type=str,
        help="semantic files (geojson or osm)",
    )
    parser.add_argument(
        "--pointcloud_files",
        nargs="*",
        type=str,
        help="pointcloud files (pcd or other pointcloud file (you should write your own handle_pointcloud))",
    )
    parser.add_argument(
        "--export",
        type=str,
        help="export to json",
    )
    args = parser.parse_args()
    workdir: str = args.workdir
    semantic_files: list[str] = args.semantic_files
    pointcloud_files: list[str] = args.pointcloud_files
    center: str | None = args.center
    export: str | None = args.export
    args = None

    index = condense(
        workdir=workdir,
        semantic_files=semantic_files,
        pointcloud_files=pointcloud_files,
        center=center,
        handle_semantic=handle_semantic,
        handle_pointcloud=handle_pointcloud,
    )
    if export:
        logger.info(f"export to {export}")
        write_json(export, index, verbose=False)
    else:
        logger.info(f"export: {json.dumps(index, indent=4)}")
        logger.warning("you can specify --export to export to json")


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
from pprint import pprint

from geocondense import CondenseOptions

from ._core import condense_geojson as condense_geojson_impl


def condense_geojson(
    *,
    input_path: str,
    output_index_path: str = None,
    output_strip_path: str = None,
    output_grids_dir: str = None,
    douglas_epsilon: float = 0.4,
    grid_h3_resolution: int = 8,
    indent: bool = False,
    sort_keys: bool = False,
    grid_features_keep_properties: bool = False,
    sparsify_h3_resolution: int = 11,
    sparsify_upper_limit: int = 48,
    debug: bool = False,
):
    # setup options
    options = CondenseOptions()
    options.douglas_epsilon = douglas_epsilon
    options.grid_h3_resolution = grid_h3_resolution
    options.indent = indent
    options.sort_keys = sort_keys
    options.grid_features_keep_properties = grid_features_keep_properties
    options.sparsify_h3_resolution = sparsify_h3_resolution
    options.sparsify_upper_limit = sparsify_upper_limit
    options.debug = debug

    # mkdir -p
    for path in [output_index_path, output_strip_path]:
        path = os.path.abspath(path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
    os.makedirs(os.path.abspath(output_grids_dir), exist_ok=True)

    # call c++ condense_geojson
    succ = condense_geojson_impl(
        input_path=input_path,
        output_index_path=output_index_path,
        output_strip_path=output_strip_path,
        output_grids_dir=output_grids_dir,
        options=options,
    )
    if not succ:
        pprint(locals())  # noqa: T203
        msg = f"failed to condense geojson: {input_path}"
        raise Exception(msg)


if __name__ == "__main__":
    import fire

    fire.core.Display = lambda lines, out: print(*lines, file=out)
    fire.Fire(condense_geojson)



from __future__ import annotations

import json
import os
import subprocess
from sys import platform

from loguru import logger


def popen(cmd: str | list[str]) -> tuple[int, str, str]:
    if isinstance(cmd, str):
        cmd = cmd.split(" ")
    p = subprocess.Popen(
        cmd,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = p.communicate()
    return p.returncode, str(out, "utf-8"), str(err, "utf-8")


def md5sum(path: str) -> str:
    assert os.path.isfile(path), f"{path} does not exist"
    if platform == "linux" or platform == "linux2":
        retcode, stdout, stderr = popen(["md5sum", path])
        assert retcode == 0
        md5 = stdout.split()[0]
    elif platform.startswith("darwin"):
        retcode, stdout, stderr = popen(["md5", path])
        assert retcode == 0
        md5 = stdout.strip().split()[-1]
    else:
        msg = "not implemented on this platform"
        raise Exception(msg)
    assert len(md5) == 32
    return md5


def read_json(path: str) -> dict:
    path = os.path.abspath(path)
    with open(path) as f:
        return json.load(f)


def write_json(path: str, data: dict, *, verbose: bool = True) -> str:
    path = os.path.abspath(path)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w", encoding="utf8") as f:
        json.dump(data, f, indent=4, ensure_ascii=False)
    if verbose:
        logger.info(f"wrote to {path}")
    return path
